import _objectSpread from '@babel/runtime/helpers/objectSpread2';
import _defineProperty from '@babel/runtime/helpers/defineProperty';
import { createScaffoldMiddleware, createAsyncMiddleware, rpcErrors, providerErrors, JRPCEngine, providerFromEngine } from '@web3auth/auth';
import { log, CHAIN_NAMESPACES, WalletInitializationError } from '@web3auth/base';
import { BaseProvider } from '@web3auth/base-provider';
import { createWalletClient, http, defineChain, createPublicClient } from 'viem';
import { createPaymasterClient, createBundlerClient, entryPoint06Address, entryPoint07Address } from 'viem/account-abstraction';
import { addHexPrefix, isHexString } from '@ethereumjs/util';
import { toBiconomySmartAccount, toEcdsaKernelSmartAccount, toNexusSmartAccount, toSafeSmartAccount, toTrustSmartAccount } from 'permissionless/accounts';

async function createAaMiddleware({
  eoaProvider,
  handlers
}) {
  const [eoaAddress] = await eoaProvider.request({
    method: "eth_accounts"
  });

  /**
   * Validates the keyholder address, and returns a normalized (i.e. lowercase)
   * copy of it.
   *
   * an error
   */
  async function validateAndNormalizeKeyholder(address, req) {
    if (typeof address === "string" && address.length > 0) {
      // ensure address is included in provided accounts
      const accounts = await handlers.getAccounts(req);
      const normalizedAccounts = accounts.map(_address => _address.toLowerCase());
      const normalizedAddress = address.toLowerCase();
      if (normalizedAccounts.includes(normalizedAddress)) {
        return normalizedAddress;
      }
    }
    throw rpcErrors.invalidParams({
      message: `Invalid parameters: must provide an Ethereum address.`
    });
  }
  async function normalizeSignSenderAddress(address, req) {
    // sender is EOA address
    if (address.toLowerCase() === eoaAddress.toLowerCase()) {
      return eoaAddress;
    }
    const [smartAccountAddress] = await handlers.getAccounts(req);
    // sender is smart account address
    if (address.toLowerCase() === smartAccountAddress.toLowerCase()) {
      // use EOA address as sender for signing
      return eoaAddress;
    }
    throw rpcErrors.invalidParams({
      message: `Invalid parameters: must provide valid sender address.`
    });
  }
  async function lookupAccounts(req, res) {
    res.result = await handlers.getAccounts(req);
  }
  async function fetchPrivateKey(req, res) {
    if (!handlers.getPrivateKey) {
      throw rpcErrors.methodNotSupported();
    }
    res.result = await handlers.getPrivateKey(req);
  }
  async function sendTransaction(req, res) {
    if (!handlers.processTransaction) {
      throw rpcErrors.methodNotSupported();
    }
    const txParams = req.params[0] || {
      from: ""
    };
    txParams.from = await validateAndNormalizeKeyholder(txParams.from, req);
    res.result = handlers.processTransaction(txParams, req);
  }
  async function signTransaction(req, res) {
    const txParams = req.params[0] || {
      from: ""
    };

    // normalize sender address
    if (txParams.from) {
      txParams.from = await normalizeSignSenderAddress(txParams.from, req);
    }
    res.result = await handlers.processSignTransaction(txParams, req);
  }
  async function ethSign(req, res) {
    let msgParams = req.params;
    const extraParams = req.params[2] || {};
    if (Array.isArray(req.params)) {
      if (!(req.params.length === 2)) throw new Error(`WalletMiddleware - incorrect params for ${req.method} method. expected [address, message]`);
      const params = req.params;
      const address = params[0];
      const message = params[1];
      msgParams = {
        from: address,
        data: message
      };
    }
    msgParams = _objectSpread(_objectSpread({}, extraParams), msgParams);
    res.result = await handlers.processEthSignMessage(msgParams, req);
  }
  async function signTypedDataV4(req, res) {
    if (!(req !== null && req !== void 0 && req.params)) throw new Error("WalletMiddleware - missing params");
    let msgParams = req.params;
    if (Array.isArray(req.params)) {
      if (!(req.params.length === 2)) throw new Error(`WalletMiddleware - incorrect params for ${req.method} method. expected [address, typedData]`);
      const params = req.params;
      const address = params[0];
      const message = params[1];
      msgParams = {
        from: address,
        data: message
      };
    }
    res.result = await handlers.processTypedMessageV4(msgParams, req);
  }
  async function personalSign(req, res) {
    let msgParams = req.params;
    const extraParams = req.params[2] || {};
    if (Array.isArray(req.params)) {
      if (!(req.params.length >= 2)) throw new Error(`WalletMiddleware - incorrect params for ${req.method} method. expected [message, address]`);
      const params = req.params;
      if (typeof params[0] === "object") {
        const {
          challenge,
          address
        } = params[0];
        msgParams = {
          from: address,
          data: challenge
        };
      } else {
        const message = params[0];
        const address = params[1];
        msgParams = {
          from: address,
          data: message
        };
      }
    }
    msgParams = _objectSpread(_objectSpread({}, extraParams), msgParams);
    res.result = await handlers.processPersonalMessage(msgParams, req);
  }
  async function fetchPublicKey(req, res) {
    if (!handlers.getPublicKey) {
      throw rpcErrors.methodNotSupported();
    }
    res.result = await handlers.getPublicKey(req);
  }
  return createScaffoldMiddleware({
    // account lookups
    eth_accounts: createAsyncMiddleware(lookupAccounts),
    eth_requestAccounts: createAsyncMiddleware(lookupAccounts),
    eth_private_key: createAsyncMiddleware(fetchPrivateKey),
    private_key: createAsyncMiddleware(fetchPrivateKey),
    eth_public_key: createAsyncMiddleware(fetchPublicKey),
    public_key: createAsyncMiddleware(fetchPublicKey),
    // tx signatures
    eth_sendTransaction: createAsyncMiddleware(sendTransaction),
    eth_signTransaction: createAsyncMiddleware(signTransaction),
    // message signatures
    eth_sign: createAsyncMiddleware(ethSign),
    eth_signTypedData_v4: createAsyncMiddleware(signTypedDataV4),
    personal_sign: createAsyncMiddleware(personalSign)
  });
}
function eoaProviderAsMiddleware(provider) {
  return async (req, res, _next, end) => {
    // send request to provider
    try {
      const providerRes = await provider.request(req);
      res.result = providerRes;
      return end();
    } catch (error) {
      return end(error);
    }
  };
}

function getProviderHandlers({
  bundlerClient,
  smartAccount,
  chain,
  eoaProvider
}) {
  const walletClient = createWalletClient({
    account: smartAccount,
    chain,
    transport: http()
  });
  return {
    getAccounts: async _ => {
      const [smartAccounts, eoaAccounts] = await Promise.all([smartAccount.getAddress(), eoaProvider.request({
        method: "eth_accounts"
      })]);
      log.info("smartAccounts", smartAccounts);
      log.info("eoaAccounts", eoaAccounts);
      return [smartAccounts, ...eoaAccounts];
    },
    getPrivateKey: async _ => {
      throw providerErrors.custom({
        message: "Smart accounts do not have private key",
        code: 4903
      });
    },
    getPublicKey: async _ => {
      throw providerErrors.custom({
        message: "Smart accounts do not have a public key. Use address instead.",
        code: 4903
      });
    },
    processTransaction: async txParams => {
      if (txParams.input && !txParams.data) txParams.data = addHexPrefix(txParams.input);
      const {
        to,
        value,
        data
      } = txParams;
      // @ts-expect-error viem types are too deep
      const userOperationParams = {
        account: smartAccount,
        calls: [{
          to,
          // Explicit conversation required to avoid value being passed as hex
          value: BigInt(value),
          data
        }]
        // should not use maxFeePerGas/maxPriorityFeePerGas from transaction params since that's fee for transaction not user operation and let bundler handle it instead
      };
      const userOpHash = await bundlerClient.sendUserOperation(userOperationParams);
      const txReceipt = await bundlerClient.waitForUserOperationReceipt({
        hash: userOpHash
      });
      if (!txReceipt.success) {
        throw providerErrors.custom({
          message: txReceipt.reason,
          code: 4905
        });
      }
      return txReceipt.receipt.transactionHash;
    },
    processSignTransaction: async txParams => {
      const {
        to,
        value,
        data
      } = txParams;
      const request = await bundlerClient.prepareUserOperation({
        account: smartAccount,
        calls: [{
          to,
          value: BigInt(value),
          data
        }]
      });
      const signature = await smartAccount.signUserOperation({
        callData: request.callData,
        callGasLimit: request.callGasLimit,
        maxFeePerGas: request.maxFeePerGas,
        maxPriorityFeePerGas: request.maxPriorityFeePerGas,
        nonce: request.nonce,
        preVerificationGas: request.preVerificationGas,
        verificationGasLimit: request.verificationGasLimit,
        signature: request.signature
      });
      return signature;
    },
    processEthSignMessage: async (_, req) => {
      return eoaProvider.request(req);
    },
    processPersonalMessage: async (msgParams, _) => {
      const message = msgParams.data;
      return walletClient.signMessage({
        account: smartAccount,
        message: isHexString(message) ? {
          raw: message
        } : message
      });
    },
    processTypedMessageV4: async (msgParams, _) => {
      try {
        const data = typeof msgParams.data === "string" ? JSON.parse(msgParams.data) : msgParams.data;
        const signature = await walletClient.signTypedData({
          account: smartAccount,
          domain: _objectSpread(_objectSpread({}, data.domain), {}, {
            verifyingContract: data.domain.verifyingContract,
            salt: data.domain.salt,
            chainId: Number(data.domain.chainId)
          }),
          primaryType: data.primaryType,
          types: data.types,
          message: data.message
        });
        return signature;
      } catch (error) {
        throw providerErrors.custom({
          message: error instanceof Error ? error.message : "Failed to sign typed data",
          code: 4905
        });
      }
    }
  };
}

var _AccountAbstractionProvider;
class AccountAbstractionProvider extends BaseProvider {
  constructor({
    config,
    state
  }) {
    super({
      config,
      state
    });
    _defineProperty(this, "PROVIDER_CHAIN_NAMESPACE", CHAIN_NAMESPACES.EIP155);
    _defineProperty(this, "_smartAccount", void 0);
    _defineProperty(this, "_publicClient", void 0);
    _defineProperty(this, "_bundlerClient", void 0);
    _defineProperty(this, "_paymasterClient", void 0);
  }
  get smartAccount() {
    return this._smartAccount;
  }
  get bundlerClient() {
    return this._bundlerClient;
  }
  get paymasterClient() {
    return this._paymasterClient;
  }
  get publicClient() {
    return this._publicClient;
  }
  async enable() {
    if (!this.state.eoaProvider) throw providerErrors.custom({
      message: "eoaProvider is not found in state, please pass it",
      code: 4902
    });
    await this.setupProvider(this.state.eoaProvider);
    return this._providerEngineProxy.request({
      method: "eth_accounts"
    });
  }
  async setupProvider(eoaProvider) {
    var _this$config$bundlerC;
    const {
      chainNamespace
    } = this.config.chainConfig;
    if (chainNamespace !== this.PROVIDER_CHAIN_NAMESPACE) throw WalletInitializationError.incompatibleChainNameSpace("Invalid chain namespace");
    const chain = defineChain({
      id: Number.parseInt(this.config.chainConfig.chainId, 16),
      // id in number form
      name: this.config.chainConfig.displayName,
      rpcUrls: {
        default: {
          http: [this.config.chainConfig.rpcTarget],
          webSocket: [this.config.chainConfig.wsTarget]
        }
      },
      blockExplorers: this.config.chainConfig.blockExplorerUrl ? {
        default: {
          name: "explorer",
          // TODO: correct name if chain config has it
          url: this.config.chainConfig.blockExplorerUrl
        }
      } : undefined,
      nativeCurrency: {
        name: this.config.chainConfig.tickerName,
        symbol: this.config.chainConfig.ticker,
        decimals: this.config.chainConfig.decimals || 18
      }
    });
    // setup public client for viem smart account
    this._publicClient = createPublicClient({
      chain,
      transport: http(this.config.chainConfig.rpcTarget)
    });
    this._smartAccount = await this.config.smartAccountInit.getSmartAccount({
      owner: eoaProvider,
      client: this._publicClient
    });

    // setup bundler and paymaster
    if (this.config.paymasterConfig) {
      var _this$config$paymaste;
      this._paymasterClient = createPaymasterClient(_objectSpread(_objectSpread({}, this.config.paymasterConfig), {}, {
        transport: (_this$config$paymaste = this.config.paymasterConfig.transport) !== null && _this$config$paymaste !== void 0 ? _this$config$paymaste : http(this.config.paymasterConfig.url)
      }));
    }
    this._bundlerClient = createBundlerClient(_objectSpread(_objectSpread({}, this.config.bundlerConfig), {}, {
      account: this.smartAccount,
      client: this._publicClient,
      transport: (_this$config$bundlerC = this.config.bundlerConfig.transport) !== null && _this$config$bundlerC !== void 0 ? _this$config$bundlerC : http(this.config.bundlerConfig.url),
      paymaster: this._paymasterClient
    }));
    const providerHandlers = getProviderHandlers({
      bundlerClient: this._bundlerClient,
      smartAccount: this._smartAccount,
      chain,
      eoaProvider
    });

    // setup rpc engine and AA middleware
    const engine = new JRPCEngine();
    const aaMiddleware = await createAaMiddleware({
      eoaProvider,
      handlers: providerHandlers
    });
    engine.push(aaMiddleware);
    const eoaMiddleware = eoaProviderAsMiddleware(eoaProvider);
    engine.push(eoaMiddleware);
    const provider = providerFromEngine(engine);
    this.updateProviderEngineProxy(provider);
    eoaProvider.once("chainChanged", () => {
      this.setupChainSwitchMiddleware();
    });
  }
  async updateAccount(_params) {
    throw providerErrors.unsupportedMethod("updateAccount. Please call it on eoaProvider");
  }
  async switchChain(_params) {
    throw providerErrors.unsupportedMethod("switchChain. Please call it on eoaProvider");
  }
  async lookupNetwork() {
    throw providerErrors.unsupportedMethod("lookupNetwork. Please call it on eoaProvider");
  }
  async setupChainSwitchMiddleware() {
    const chainConfig = await this.state.eoaProvider.request({
      method: "eth_provider_config"
    });
    this.update({
      chainId: chainConfig.chainId
    });
    this.configure({
      chainConfig: _objectSpread(_objectSpread({}, chainConfig), {}, {
        chainNamespace: CHAIN_NAMESPACES.EIP155,
        chainId: chainConfig.chainId,
        rpcTarget: chainConfig.rpcTarget
      })
    });
    return this.setupProvider(this.state.eoaProvider);
  }
}
_AccountAbstractionProvider = AccountAbstractionProvider;
_defineProperty(AccountAbstractionProvider, "getProviderInstance", async params => {
  const providerFactory = new _AccountAbstractionProvider({
    config: params
  });
  await providerFactory.setupProvider(params.eoaProvider);
  providerFactory.update({
    eoaProvider: params.eoaProvider
  });
  return providerFactory;
});

const SMART_ACCOUNT = {
  BICONOMY: "biconomy",
  NEXUS: "nexus",
  KERNEL: "kernel",
  LIGHT: "light",
  SAFE: "safe",
  SIMPLE: "simple",
  TRUST: "trust"
};

class BiconomySmartAccount {
  constructor(options) {
    _defineProperty(this, "name", SMART_ACCOUNT.BICONOMY);
    _defineProperty(this, "options", void 0);
    this.options = options;
  }
  async getSmartAccount(params) {
    var _this$options, _this$options2;
    return toBiconomySmartAccount(_objectSpread(_objectSpread(_objectSpread({}, this.options || {}), {}, {
      entryPoint: {
        address: ((_this$options = this.options) === null || _this$options === void 0 || (_this$options = _this$options.entryPoint) === null || _this$options === void 0 ? void 0 : _this$options.address) || entryPoint06Address,
        version: ((_this$options2 = this.options) === null || _this$options2 === void 0 || (_this$options2 = _this$options2.entryPoint) === null || _this$options2 === void 0 ? void 0 : _this$options2.version) || "0.6"
      }
    }, params), {}, {
      owners: [params.owner],
      client: params.client
    }));
  }
}

class KernelSmartAccount {
  constructor(options) {
    _defineProperty(this, "name", SMART_ACCOUNT.KERNEL);
    _defineProperty(this, "options", void 0);
    this.options = options;
  }
  async getSmartAccount(params) {
    return toEcdsaKernelSmartAccount(_objectSpread(_objectSpread(_objectSpread({}, this.options || {}), params), {}, {
      owners: [params.owner],
      client: params.client
    }));
  }
}

class NexusSmartAccount {
  constructor(options) {
    _defineProperty(this, "name", SMART_ACCOUNT.NEXUS);
    _defineProperty(this, "options", void 0);
    this.options = options;
  }
  async getSmartAccount(params) {
    var _this$options, _this$options2, _this$options3;
    return toNexusSmartAccount(_objectSpread(_objectSpread(_objectSpread({}, this.options || {}), {}, {
      entryPoint: {
        address: ((_this$options = this.options) === null || _this$options === void 0 || (_this$options = _this$options.entryPoint) === null || _this$options === void 0 ? void 0 : _this$options.address) || entryPoint07Address,
        version: ((_this$options2 = this.options) === null || _this$options2 === void 0 || (_this$options2 = _this$options2.entryPoint) === null || _this$options2 === void 0 ? void 0 : _this$options2.version) || "0.7"
      },
      version: ((_this$options3 = this.options) === null || _this$options3 === void 0 ? void 0 : _this$options3.version) || "1.0.0"
    }, params), {}, {
      owners: [params.owner],
      client: params.client
    }));
  }
}

class SafeSmartAccount {
  constructor(options) {
    _defineProperty(this, "name", SMART_ACCOUNT.SAFE);
    _defineProperty(this, "options", void 0);
    this.options = options;
  }
  async getSmartAccount(params) {
    var _this$options, _this$options2, _this$options3;
    return toSafeSmartAccount(_objectSpread(_objectSpread(_objectSpread({}, this.options || {}), {}, {
      entryPoint: {
        address: ((_this$options = this.options) === null || _this$options === void 0 || (_this$options = _this$options.entryPoint) === null || _this$options === void 0 ? void 0 : _this$options.address) || entryPoint07Address,
        version: ((_this$options2 = this.options) === null || _this$options2 === void 0 || (_this$options2 = _this$options2.entryPoint) === null || _this$options2 === void 0 ? void 0 : _this$options2.version) || "0.7"
      },
      version: ((_this$options3 = this.options) === null || _this$options3 === void 0 ? void 0 : _this$options3.version) || "1.4.1"
    }, params), {}, {
      owners: [params.owner],
      client: params.client
    }));
  }
}

class TrustSmartAccount {
  constructor(options) {
    _defineProperty(this, "name", SMART_ACCOUNT.TRUST);
    _defineProperty(this, "options", void 0);
    this.options = options;
  }
  async getSmartAccount(params) {
    var _this$options, _this$options2;
    return toTrustSmartAccount(_objectSpread(_objectSpread(_objectSpread({}, this.options || {}), {}, {
      entryPoint: {
        address: ((_this$options = this.options) === null || _this$options === void 0 || (_this$options = _this$options.entryPoint) === null || _this$options === void 0 ? void 0 : _this$options.address) || entryPoint06Address,
        version: ((_this$options2 = this.options) === null || _this$options2 === void 0 || (_this$options2 = _this$options2.entryPoint) === null || _this$options2 === void 0 ? void 0 : _this$options2.version) || "0.6"
      }
    }, params), {}, {
      owner: params.owner,
      client: params.client
    }));
  }
}

export { AccountAbstractionProvider, BiconomySmartAccount, KernelSmartAccount, NexusSmartAccount, SafeSmartAccount, TrustSmartAccount, getProviderHandlers };
