'use strict';

var _objectSpread = require('@babel/runtime/helpers/objectSpread2');
var _defineProperty = require('@babel/runtime/helpers/defineProperty');
var auth = require('@web3auth/auth');
var base = require('@web3auth/base');
var baseProvider = require('@web3auth/base-provider');
var viem = require('viem');
var accountAbstraction = require('viem/account-abstraction');
var ethRpcMiddlewares = require('../rpc/ethRpcMiddlewares.js');
var utils = require('./utils.js');

var _AccountAbstractionProvider;
class AccountAbstractionProvider extends baseProvider.BaseProvider {
  constructor({
    config,
    state
  }) {
    super({
      config,
      state
    });
    _defineProperty(this, "PROVIDER_CHAIN_NAMESPACE", base.CHAIN_NAMESPACES.EIP155);
    _defineProperty(this, "_smartAccount", void 0);
    _defineProperty(this, "_publicClient", void 0);
    _defineProperty(this, "_bundlerClient", void 0);
    _defineProperty(this, "_paymasterClient", void 0);
  }
  get smartAccount() {
    return this._smartAccount;
  }
  get bundlerClient() {
    return this._bundlerClient;
  }
  get paymasterClient() {
    return this._paymasterClient;
  }
  get publicClient() {
    return this._publicClient;
  }
  async enable() {
    if (!this.state.eoaProvider) throw auth.providerErrors.custom({
      message: "eoaProvider is not found in state, please pass it",
      code: 4902
    });
    await this.setupProvider(this.state.eoaProvider);
    return this._providerEngineProxy.request({
      method: "eth_accounts"
    });
  }
  async setupProvider(eoaProvider) {
    var _this$config$bundlerC;
    const {
      chainNamespace
    } = this.config.chainConfig;
    if (chainNamespace !== this.PROVIDER_CHAIN_NAMESPACE) throw base.WalletInitializationError.incompatibleChainNameSpace("Invalid chain namespace");
    const chain = viem.defineChain({
      id: Number.parseInt(this.config.chainConfig.chainId, 16),
      // id in number form
      name: this.config.chainConfig.displayName,
      rpcUrls: {
        default: {
          http: [this.config.chainConfig.rpcTarget],
          webSocket: [this.config.chainConfig.wsTarget]
        }
      },
      blockExplorers: this.config.chainConfig.blockExplorerUrl ? {
        default: {
          name: "explorer",
          // TODO: correct name if chain config has it
          url: this.config.chainConfig.blockExplorerUrl
        }
      } : undefined,
      nativeCurrency: {
        name: this.config.chainConfig.tickerName,
        symbol: this.config.chainConfig.ticker,
        decimals: this.config.chainConfig.decimals || 18
      }
    });
    // setup public client for viem smart account
    this._publicClient = viem.createPublicClient({
      chain,
      transport: viem.http(this.config.chainConfig.rpcTarget)
    });
    this._smartAccount = await this.config.smartAccountInit.getSmartAccount({
      owner: eoaProvider,
      client: this._publicClient
    });

    // setup bundler and paymaster
    if (this.config.paymasterConfig) {
      var _this$config$paymaste;
      this._paymasterClient = accountAbstraction.createPaymasterClient(_objectSpread(_objectSpread({}, this.config.paymasterConfig), {}, {
        transport: (_this$config$paymaste = this.config.paymasterConfig.transport) !== null && _this$config$paymaste !== void 0 ? _this$config$paymaste : viem.http(this.config.paymasterConfig.url)
      }));
    }
    this._bundlerClient = accountAbstraction.createBundlerClient(_objectSpread(_objectSpread({}, this.config.bundlerConfig), {}, {
      account: this.smartAccount,
      client: this._publicClient,
      transport: (_this$config$bundlerC = this.config.bundlerConfig.transport) !== null && _this$config$bundlerC !== void 0 ? _this$config$bundlerC : viem.http(this.config.bundlerConfig.url),
      paymaster: this._paymasterClient
    }));
    const providerHandlers = utils.getProviderHandlers({
      bundlerClient: this._bundlerClient,
      smartAccount: this._smartAccount,
      chain,
      eoaProvider
    });

    // setup rpc engine and AA middleware
    const engine = new auth.JRPCEngine();
    const aaMiddleware = await ethRpcMiddlewares.createAaMiddleware({
      eoaProvider,
      handlers: providerHandlers
    });
    engine.push(aaMiddleware);
    const eoaMiddleware = ethRpcMiddlewares.eoaProviderAsMiddleware(eoaProvider);
    engine.push(eoaMiddleware);
    const provider = auth.providerFromEngine(engine);
    this.updateProviderEngineProxy(provider);
    eoaProvider.once("chainChanged", () => {
      this.setupChainSwitchMiddleware();
    });
  }
  async updateAccount(_params) {
    throw auth.providerErrors.unsupportedMethod("updateAccount. Please call it on eoaProvider");
  }
  async switchChain(_params) {
    throw auth.providerErrors.unsupportedMethod("switchChain. Please call it on eoaProvider");
  }
  async lookupNetwork() {
    throw auth.providerErrors.unsupportedMethod("lookupNetwork. Please call it on eoaProvider");
  }
  async setupChainSwitchMiddleware() {
    const chainConfig = await this.state.eoaProvider.request({
      method: "eth_provider_config"
    });
    this.update({
      chainId: chainConfig.chainId
    });
    this.configure({
      chainConfig: _objectSpread(_objectSpread({}, chainConfig), {}, {
        chainNamespace: base.CHAIN_NAMESPACES.EIP155,
        chainId: chainConfig.chainId,
        rpcTarget: chainConfig.rpcTarget
      })
    });
    return this.setupProvider(this.state.eoaProvider);
  }
}
_AccountAbstractionProvider = AccountAbstractionProvider;
_defineProperty(AccountAbstractionProvider, "getProviderInstance", async params => {
  const providerFactory = new _AccountAbstractionProvider({
    config: params
  });
  await providerFactory.setupProvider(params.eoaProvider);
  providerFactory.update({
    eoaProvider: params.eoaProvider
  });
  return providerFactory;
});

exports.AccountAbstractionProvider = AccountAbstractionProvider;
