import _objectSpread from '@babel/runtime/helpers/objectSpread2';
import _defineProperty from '@babel/runtime/helpers/defineProperty';
import { providerErrors, JRPCEngine, providerFromEngine } from '@web3auth/auth';
import { CHAIN_NAMESPACES, WalletInitializationError } from '@web3auth/base';
import { BaseProvider } from '@web3auth/base-provider';
import { defineChain, createPublicClient, http } from 'viem';
import { createPaymasterClient, createBundlerClient } from 'viem/account-abstraction';
import { createAaMiddleware, eoaProviderAsMiddleware } from '../rpc/ethRpcMiddlewares.js';
import { getProviderHandlers } from './utils.js';

var _AccountAbstractionProvider;
class AccountAbstractionProvider extends BaseProvider {
  constructor({
    config,
    state
  }) {
    super({
      config,
      state
    });
    _defineProperty(this, "PROVIDER_CHAIN_NAMESPACE", CHAIN_NAMESPACES.EIP155);
    _defineProperty(this, "_smartAccount", void 0);
    _defineProperty(this, "_publicClient", void 0);
    _defineProperty(this, "_bundlerClient", void 0);
    _defineProperty(this, "_paymasterClient", void 0);
  }
  get smartAccount() {
    return this._smartAccount;
  }
  get bundlerClient() {
    return this._bundlerClient;
  }
  get paymasterClient() {
    return this._paymasterClient;
  }
  get publicClient() {
    return this._publicClient;
  }
  async enable() {
    if (!this.state.eoaProvider) throw providerErrors.custom({
      message: "eoaProvider is not found in state, please pass it",
      code: 4902
    });
    await this.setupProvider(this.state.eoaProvider);
    return this._providerEngineProxy.request({
      method: "eth_accounts"
    });
  }
  async setupProvider(eoaProvider) {
    var _this$config$bundlerC;
    const {
      chainNamespace
    } = this.config.chainConfig;
    if (chainNamespace !== this.PROVIDER_CHAIN_NAMESPACE) throw WalletInitializationError.incompatibleChainNameSpace("Invalid chain namespace");
    const chain = defineChain({
      id: Number.parseInt(this.config.chainConfig.chainId, 16),
      // id in number form
      name: this.config.chainConfig.displayName,
      rpcUrls: {
        default: {
          http: [this.config.chainConfig.rpcTarget],
          webSocket: [this.config.chainConfig.wsTarget]
        }
      },
      blockExplorers: this.config.chainConfig.blockExplorerUrl ? {
        default: {
          name: "explorer",
          // TODO: correct name if chain config has it
          url: this.config.chainConfig.blockExplorerUrl
        }
      } : undefined,
      nativeCurrency: {
        name: this.config.chainConfig.tickerName,
        symbol: this.config.chainConfig.ticker,
        decimals: this.config.chainConfig.decimals || 18
      }
    });
    // setup public client for viem smart account
    this._publicClient = createPublicClient({
      chain,
      transport: http(this.config.chainConfig.rpcTarget)
    });
    this._smartAccount = await this.config.smartAccountInit.getSmartAccount({
      owner: eoaProvider,
      client: this._publicClient
    });

    // setup bundler and paymaster
    if (this.config.paymasterConfig) {
      var _this$config$paymaste;
      this._paymasterClient = createPaymasterClient(_objectSpread(_objectSpread({}, this.config.paymasterConfig), {}, {
        transport: (_this$config$paymaste = this.config.paymasterConfig.transport) !== null && _this$config$paymaste !== void 0 ? _this$config$paymaste : http(this.config.paymasterConfig.url)
      }));
    }
    this._bundlerClient = createBundlerClient(_objectSpread(_objectSpread({}, this.config.bundlerConfig), {}, {
      account: this.smartAccount,
      client: this._publicClient,
      transport: (_this$config$bundlerC = this.config.bundlerConfig.transport) !== null && _this$config$bundlerC !== void 0 ? _this$config$bundlerC : http(this.config.bundlerConfig.url),
      paymaster: this._paymasterClient
    }));
    const providerHandlers = getProviderHandlers({
      bundlerClient: this._bundlerClient,
      smartAccount: this._smartAccount,
      chain,
      eoaProvider
    });

    // setup rpc engine and AA middleware
    const engine = new JRPCEngine();
    const aaMiddleware = await createAaMiddleware({
      eoaProvider,
      handlers: providerHandlers
    });
    engine.push(aaMiddleware);
    const eoaMiddleware = eoaProviderAsMiddleware(eoaProvider);
    engine.push(eoaMiddleware);
    const provider = providerFromEngine(engine);
    this.updateProviderEngineProxy(provider);
    eoaProvider.once("chainChanged", () => {
      this.setupChainSwitchMiddleware();
    });
  }
  async updateAccount(_params) {
    throw providerErrors.unsupportedMethod("updateAccount. Please call it on eoaProvider");
  }
  async switchChain(_params) {
    throw providerErrors.unsupportedMethod("switchChain. Please call it on eoaProvider");
  }
  async lookupNetwork() {
    throw providerErrors.unsupportedMethod("lookupNetwork. Please call it on eoaProvider");
  }
  async setupChainSwitchMiddleware() {
    const chainConfig = await this.state.eoaProvider.request({
      method: "eth_provider_config"
    });
    this.update({
      chainId: chainConfig.chainId
    });
    this.configure({
      chainConfig: _objectSpread(_objectSpread({}, chainConfig), {}, {
        chainNamespace: CHAIN_NAMESPACES.EIP155,
        chainId: chainConfig.chainId,
        rpcTarget: chainConfig.rpcTarget
      })
    });
    return this.setupProvider(this.state.eoaProvider);
  }
}
_AccountAbstractionProvider = AccountAbstractionProvider;
_defineProperty(AccountAbstractionProvider, "getProviderInstance", async params => {
  const providerFactory = new _AccountAbstractionProvider({
    config: params
  });
  await providerFactory.setupProvider(params.eoaProvider);
  providerFactory.update({
    eoaProvider: params.eoaProvider
  });
  return providerFactory;
});

export { AccountAbstractionProvider };
